# -*- coding: utf-8 -*-


class DTShellConstants(object):
    commands_remote_url = 'https://github.com/duckietown/duckietown-shell-commands'
    ROOT = '~/.dt-shell/'
    ENV_COMMANDS = 'DTSHELL_COMMANDS'

    DT1_TOKEN_CONFIG_KEY = 'token_dt1'
    CONFIG_DOCKER_USERNAME = 'docker_username'



# -*- coding: utf-8 -*-
from __future__ import print_function

import datetime
import json
import os
import urllib2

import ruamel
import ruamel.yaml as yaml
import termcolor
from system_cmd import system_cmd_result
from whichcraft import which

from . import __version__, dtslogger
from .constants import DTShellConstants


class CouldNotCheck(Exception):
    pass


def get_last_version_fresh():
    url = 'https://pypi.org/pypi/duckietown-shell/json'

    req = urllib2.Request(url)
    try:
        res = urllib2.urlopen(req, timeout=1)
        data = res.read()

    except urllib2.URLError as e:
        # msg = 'Cannot connect to server %s: %s' % (url, (e))
        # raise Exception(msg)
        if which('curl') is not None:
            res = system_cmd_result('.', ['curl', url])
            data = res.stdout
        else:
            raise CouldNotCheck()

    info = json.loads(data)
    last_version = info['info']['version']
    return last_version


def get_cache_filename():
    d0 = os.path.expanduser(DTShellConstants.ROOT)
    return os.path.join(d0, 'pypi-cache.yaml')


class NoCacheAvailable(Exception):
    pass


def read_cache():
    try:
        fn = get_cache_filename()
        if os.path.exists(fn):
            data = open(fn).read()
            interpreted = yaml.load(data, Loader=ruamel.yaml.Loader)
            version = interpreted['version']
            dt = interpreted['timestamp']
            return version, dt
        else:
            msg = 'File %s does not exist.' % fn
            raise NoCacheAvailable(msg)
    except Exception as e:
        msg = 'Could not read cache: %s' % e
        raise NoCacheAvailable(msg)


def write_cache(version, dt):
    fn = get_cache_filename()
    d0 = os.path.dirname(fn)
    if not os.path.exists(d0):
        os.makedirs(d0)

    with open(fn, 'w') as f:
        contents = dict(version=version, timestamp=dt)
        y = yaml.dump(contents)
        f.write(y)


def get_last_version():
    now = datetime.datetime.now()
    update = False
    try:
        version, timestamp = read_cache()
    except NoCacheAvailable:
        timestamp = now
        update = True

    if not update:
        delta = now - timestamp
        print(delta)
        if delta > datetime.timedelta(hours=1):
            dtslogger.debug('Cache is outdated.')
            update = True

    if update:
        dtslogger.debug('Getting last version from PyPI.')
        version = get_last_version_fresh()
        write_cache(version, now)

    return version


def check_if_outdated():
    latest_version = get_last_version()
    # print('last version: %r' % latest_version)
    # print('installed: %r' % __version__)
    if __version__ != latest_version:
        msg = 'There is an updated duckietown-shell available, version %s (you have %s).' % (
            latest_version, __version__)
        msg += '\n\nPlease run:\n\npip update -U --no-cache-dir duckietown-shell==%s' % (latest_version)
        msg += '\n\n'
        print(termcolor.colored(msg, 'yellow'))



import json
import os
import urllib2

import dateutil.parser
from contracts import raise_wrapped, indent

from . import dtslogger


class Storage(object):
    done = False


def get_duckietown_server_url():
    V = 'DTSERVER'
    DEFAULT = 'https://challenges.duckietown.org/v2'
    if V in os.environ:
        use = os.environ[V]
        if not Storage.done:
            msg = 'Using server %s instead of default %s' % (use, DEFAULT)
            dtslogger.info(msg)
            Storage.done = True
        return use
    else:
        return DEFAULT


class RequestException(Exception):
    pass


class ConnectionError(RequestException):
    """ The server could not be reached or completed request or
        provided an invalid or not well-formatted answer. """


class RequestFailed(RequestException):
    """
        The server said the request was invalid.

        Answered  {'ok': False, 'error': msg}
    """


def make_server_request(token, endpoint, data=None, method='GET', timeout=3):
    """
        Raise RequestFailed or ConnectionError.

        Returns the result in 'result'.
    """
    server = get_duckietown_server_url()
    url = server + endpoint

    headers = {'X-Messaging-Token': token}
    if data is not None:
        data = json.dumps(data)
    req = urllib2.Request(url, headers=headers, data=data)
    req.get_method = lambda: method
    try:
        res = urllib2.urlopen(req, timeout=timeout)
        data = res.read()
    except urllib2.URLError as e:
        msg = 'Cannot connect to server %s' % url
        raise_wrapped(ConnectionError, e, msg)
        raise

    try:
        result = json.loads(data)
    except ValueError as e:
        msg = 'Cannot read answer from server.'
        msg += '\n\n' + indent(data, '  > ')
        raise_wrapped(ConnectionError, e, msg)
        raise

    if not isinstance(result, dict) or 'ok' not in result:
        msg = 'Server provided invalid JSON response. Expected a dict with "ok" in it.'
        msg += '\n\n' + indent(data, '  > ')
        raise ConnectionError(msg)

    if result['ok']:
        if 'result' not in result:
            msg = 'Server provided invalid JSON response. Expected a field "result".'
            msg += '\n\n' + indent(result, '  > ')
            raise ConnectionError(msg)
        return result['result']
    else:
        msg = 'Failed request for %s:\n%s' % (url, result.get('error', result))
        raise RequestFailed(msg)


def get_dtserver_user_info(token):
    """ Returns a dictionary with information about the user """
    endpoint = '/info'
    method = 'GET'
    data = None
    return make_server_request(token, endpoint, data=data, method=method)


def dtserver_update_challenge(token, queue, challenge_parameters):
    endpoint = '/challenge-update'
    method = 'POST'
    data = {'queue': queue, 'challenge_parameters': challenge_parameters}
    return make_server_request(token, endpoint, data=data, method=method)


def dtserver_submit(token, queue, data):
    endpoint = '/submissions'
    method = 'POST'
    data = {'queue': queue, 'parameters': data}
    return make_server_request(token, endpoint, data=data, method=method)


def dtserver_retire(token, submission_id):
    endpoint = '/submissions'
    method = 'DELETE'
    data = {'submission_id': submission_id}
    return make_server_request(token, endpoint, data=data, method=method)


def dtserver_get_user_submissions(token):
    """ Returns a dictionary with information about the user submissions """
    endpoint = '/submissions'
    method = 'GET'
    data = {}
    submissions = make_server_request(token, endpoint, data=data, method=method)

    for v in submissions.values():
        for k in ['date_submitted', 'last_status_change']:
            v[k] = dateutil.parser.parse(v[k])
    return submissions


def dtserver_work_submission(token, submission_id, machine_id, process_id, evaluator_version):
    endpoint = '/take-submission'
    method = 'GET'
    data = {'submission_id': submission_id,
            'machine_id': machine_id,
            'process_id': process_id,
            'evaluator_version': evaluator_version}
    return make_server_request(token, endpoint, data=data, method=method)


def dtserver_report_job(token, job_id, result, stats, machine_id,
                        process_id, evaluation_container, evaluator_version):
    endpoint = '/take-submission'
    method = 'POST'
    data = {'job_id': job_id,
            'result': result,
            'stats': stats,
            'machine_id': machine_id,
            'process_id': process_id,
            'evaluation_container': evaluation_container,
            'evaluator_version': evaluator_version,
            }
    return make_server_request(token, endpoint, data=data, method=method)



import getpass
import sys

from system_cmd import system_cmd_result, CmdException
from whichcraft import which

from dt_shell.constants import DTShellConstants


class InvalidEnvironment(Exception):
    pass


def check_docker_environment():
    print('checking docker environment')
    check_executable_exists('docker')

    if on_linux():
        username = getpass.getuser()
        if username != 'root':
            check_user_in_group('docker')
        print('checked groups')
    else:
        print('skipping env check')


    try:
        import docker
    except Exception as e:
        msg ='Could not import package docker:\n%s' % e
        msg += '\n\nTry    pip install --user -U docker'
        raise InvalidEnvironment(msg)

    try:
        client = docker.from_env()
        containers = client.containers.list(filters=dict(status='running'))
    except Exception as e:
        msg = 'I cannot communicate with Docker:\n%s' % e
        msg += '\n\nMake sure the docker service is running.'
        raise InvalidEnvironment(msg)



def on_linux():
    return sys.platform.startswith('linux')


def check_executable_exists(cmdname):
    p = which(cmdname)
    if p is None:
        msg = 'Could not find executable "%s".' % cmdname
        raise InvalidEnvironment(msg)


def check_user_in_group(name):
    active_groups = get_active_groups(username=None)

    if name not in active_groups:
        msg = 'The user is not in group "%s".' % name
        msg += '\n\nIt belongs to groups: %s.' % ", ".join(sorted(active_groups))

        msg += '\n\nNote that when you add a user to a group, you need to login in and out.'
        raise InvalidEnvironment(msg)


def check_git_supports_superproject():
    res = system_cmd_result('.', ['git', '--version'],
                            display_stdout=False,
                            display_stderr=False,
                            raise_on_error=True,
                            capture_keyboard_interrupt=False,
                            env=None)


def get_active_groups(username=None):
    cmd = ['groups']

    if username:
        cmd.append(username)

    try:
        res = system_cmd_result('.', cmd,
                                display_stdout=False,
                                display_stderr=False,
                                raise_on_error=True,
                                capture_keyboard_interrupt=False,
                                env=None)
    except CmdException as e:
        raise Exception(str(e))
    active_groups = res.stdout.split()  # XXX
    return active_groups


def get_dockerhub_username(shell=None):
    if shell is None:
        from .cli import DTShell
        shell = DTShell()
    k = DTShellConstants.CONFIG_DOCKER_USERNAME
    if k not in shell.config:
        msg = 'Please set docker username using\n\n dts challenges config --docker-username <USERNAME>'
        raise Exception(msg)

    username = shell.config[k]
    return username



# -*- coding: utf-8 -*-
import logging
import sys
import traceback

logging.basicConfig()

dtslogger = logging.getLogger('dts')
dtslogger.setLevel(logging.INFO)

import termcolor

__version__ = '0.2.31'

from .cli import DTShell

from .dt_command_abs import DTCommandAbs
from .dt_command_placeholder import DTCommandPlaceholder

dtslogger.info('duckietown-shell %s' % __version__)


def cli_main():
    # TODO: register handler for Ctrl-C

    from dt_shell.env_checks import InvalidEnvironment

    shell = DTShell()
    arguments = sys.argv[1:]

    known_exceptions = (InvalidEnvironment,)

    try:
        if arguments:
            cmdline = " ".join(arguments)
            shell.onecmd(cmdline)
        else:
            shell.cmdloop()
    except known_exceptions as e:
        msg = str(e)
        termcolor.cprint(msg, 'yellow')
        sys.exit(1)
    except Exception as e:
        msg = traceback.format_exc(e)
        termcolor.cprint(msg, 'red')
        sys.exit(2)
